/*
 *	analy.h
 */

/*	Copyright (C) 1999, 2000, 2001 Sebastian Biallas (sb@web-productions.de)
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License version 2 as
 *	published by the Free Software Foundation.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef analy_h
#define analy_h

#include <stdio.h>
#include "asm.h"
#include "global.h"
#include "common.h"
#include "codeanaly.h"
#include "dataanaly.h"
#include "stddata.h"

extern int num_ops_parsed;

//#define ANALY_TIMINGS

class analyser;

/*
 *	these are the different possibilities of a branch
 *	to support further processors other types can be added
 */
typedef enum {
			brnobranch,						// straight exec. flow
			brjump,
			brreturn,
			brcall,
			brjXX
} tbranchtype;

/*
 *   internal opcodes are interchanged in this format
 */
#define OPCODE dis_insn

/*
 *
 */
class analy_disassembler: public object {
public:
	analyser			*analy;
	disassembler        *disasm;
			void			init(analyser *A);
			int 			load(ht_object_stream *f);
	virtual	void			done();

	virtual	ADDR			branchaddr(OPCODE *opcode) = 0;
	virtual	void			init_disasm();
	virtual	tbranchtype 	isbranch(OPCODE *opcode) = 0;
	virtual	void			store(ht_object_stream *f);
};

/***************************************************************************/

typedef enum {xrefread, xrefwrite, xrefoffset, xrefjump, xrefcall} txreftype;

struct txref {
	txref	*next;
	ADDR		addr;
	txreftype	type;
};

struct tcomment {
	tcomment	*next;
	char		*text;
};

struct tlabel;

struct taddr {
	// the address
	ADDR		addr;
	// this is a tree structure (key is addr)
	taddr	*left, *right;
	// attached label
	tlabel	*label;
	// attached xrefs
	txref	*xreflist;
	// attached comments
	tcomment	*comments;
	// for data types
	taddrtype	type;
	// the function the address belongs to (if applicable)
	taddr	*thisfunc;
	// some flags
	int		flags;
};

/*
 * taddr.flags:
 */
#define AF_DELETED 1
#define AF_FUNCTION_SET 2
#define AF_FUNCTION_END 4

typedef enum {scvalid, scread, scwrite,	screadwrite, sccode, scinitialized} tsectype;

typedef enum {	acread, acwrite, acoffset } taccesstype;

struct taccess	{
		bool			indexed;
		bool			write;
		taccesstype 	type;
};

typedef enum { label_unknown, label_func, label_loc, label_data } labeltype;

struct tlabel {
	bool 	freed;
	labeltype	type;
	taddr	*addr;
	char		*name;
	tlabel	*left, *right;
};

struct addrqueueitem {
	ADDR		addr;
	ADDR		func;
};


class codeanalyser;
class dataanalyser;


class analyser: public object	{
public:
	ADDR			addr;
	queue		*addr_queue;
	int			ops_parsed;							// for continueing
	bool			active;
	ADDR			next_explored,	first_explored, last_explored;
	area			*explored;
	taddr		*addrs;
	codeanalyser	*code;
	dataanalyser	*data;
	analy_disassembler	*analy_disasm;
	disassembler	*disasm;
	tlabel		*labels;
	int			addrthreshold, labelthreshold;
	int			cur_addr_ops, cur_label_ops;                 // for threshold
	int			maxopcodelength;
	taddr          *cur_func;
#ifdef ANALY_TIMINGS
	int			timer;
#endif

			void			init();
			int			load(ht_object_stream *f);
	virtual	void			done();

			bool			addaddrlabel(ADDR Addr, char *Prefix, labeltype type, taddr *infunc=NULL);
			void	 		addcomment(ADDR Addr, int line, char *c);
			bool			addlabel(ADDR Addr, char *label, labeltype type, taddr *infunc=NULL);
			void			addxref(ADDR from, ADDR to, txreftype action);
			void	 		assigncomment(ADDR Addr, int line, char *c);
			bool			assignlabel(ADDR Addr, char *label, labeltype type, taddr *infunc=NULL);
	virtual	void			beginanalysis(ADDR Addr);
	virtual	void			*bufptr(ADDR Addr) = 0;
			bool	  		continueanalysis();
			void			continueanalysisat(ADDR Addr);
			void			dataaccess(ADDR Addr, taccess	access);
			void			deleteaddr(ADDR Addr);
			void			deletelabel(ADDR Addr);
			void			disablelabel(tlabel *label);
			void			dobranch(tbranchtype branch, OPCODE *opcode, int len);
			void			engage_codeanalyser();
			taddr		*enum_addrs(ADDR Addr);
			taddr		*enum_addrs_back(ADDR Addr);
			tlabel		*enum_labels(char *at);
			tlabel		*enum_labels_back(char *at);
	virtual	taddrtypetype 	examinedata(ADDR Addr);
	virtual	void			examineopcode(OPCODE *opcode) = 0;
	virtual	FILEADDR		fileaddr(ADDR Addr) = 0;
			taddr		*findaddr(ADDR Addr);
			taddr		*findaddrfunc(ADDR Addr);
			taddr		*findaddrlabel(ADDR Addr);
			tlabel		*findlabel(char *label);
			void			finish();
			void			freeaddr(taddr *Addr);
			void			freeaddrs(taddr *addrs);
			void			freelabel(tlabel *label);
			void			freelabels(tlabel *labels);
			int			getaddrcount();
			int			getlabelcount();
			tlabel 		*getaddrlabel(ADDR Addr);
	virtual	char			*getaddrsectionname(ADDR Addr);
			bool			gotoaddr(ADDR Addr, ADDR func);
	virtual	void 		init_codeanalyser();
	virtual	void			init_dataanalyser();
	virtual	void			init_unasm() = 0;
	virtual	void			log(char *s);						// stub
	virtual	CPU_ADDR 		mapaddr(ADDR Addr);					// stub
			taddr		*newaddr(ADDR Addr);
			tlabel		*newlabel(char *label, taddr *Addr, labeltype type, taddr *infunc);
//			tlabel		*newlabel(char *label, taddr *Addr, labeltype type);
	virtual	ADDR			nextvalid(ADDR Addr) = 0;
	virtual	void			notifyprogress(ADDR	Addr);			// stub
			void			optimize_addr_tree();
			void			optimize_label_tree();
			bool			popaddr(ADDR *Addr, ADDR *func);
			void			pushaddr(ADDR Addr, ADDR func);
	virtual	int			query_config(int mode);				// stub
			bool			savetychangeaddr(ADDR Addr);
			void			setactive(bool mode);
			void			setaddrfunc(taddr *a, taddr *func);
			void			setdisasm(disassembler *Disasm);
			void			set_addr_tree_optimize_threshold(int threshold);
			void			set_label_tree_optimize_threshold(int threshold);
	virtual	void			store(ht_object_stream *f);
	virtual	bool			validaddr(ADDR	Addr, tsectype	action) = 0;
			bool			validcodeaddr(ADDR Addr);
			bool			validreadaddr(ADDR Addr);
			bool			validwriteaddr(ADDR	Addr);

///  interface only (there's no internal use)
	int		mode;

	virtual   assembler      *create_assembler();
			tcomment		*getcomments(ADDR Addr);
			char			*getdisasmstr(ADDR Addr);
			char			*getdisasmstr_formatted(ADDR Addr);
			bool			getformattedline(ADDR Addr, int line, char *buf, int *length);
			bool			getformattedline_length(ADDR Addr, int line, int *length);
	virtual   char			*getname();
			txref		*getxrefs(ADDR Addr);
			void			setdisplaymode(int enable, int disable);
			void			toggledisplaymode(int toggle);
	virtual	ADDR			vaddr(FILEADDR fileaddr);
};

/* display modes */
#define ANALY_EDIT_BYTES 1
#define ANALY_TRANSLATE_SYMBOLS 2
#define ANALY_COLLAPSE_XREFS 4

/* query_config() constants */
#define Q_DO_ANALYSIS 1
#define Q_ENGAGE_CODE_ANALYSER 2
#define Q_ENGAGE_DATA_ANALYSER 3

/* interesting constants */
#define INVALID_ADDR ((ADDR)-1)
#define EXTERNAL_ENTRY (((ADDR)-1)-1)
#define INVALID_FILE_OFS ((dword)-1)

/* analyser system constants */
#define MAX_OPS_PER_CONTINUE 10

/*
 * test;
 */
extern analyser *testanaly;

#endif
