/* 
 *	HT Editor
 *	cstream.h
 *
 *	Copyright (C) 1999, 2000, 2001 Stefan Weyergraf (stefan@weyergraf.de)
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License version 2 as
 *	published by the Free Software Foundation.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef __CSTREAM_H__
#define __CSTREAM_H__

#include "htdata.h"
#include "stream.h"

/*
 *	CLASS ht_cache_page
 */

class ht_cache_page: public ht_data {
public:
	dword read_count;
	dword write_count;
	dword alloc_index;
	FILEOFS offset;
	dword size;
	byte *data;

			ht_cache_page(FILEOFS offset, dword size, dword alloc_index);
			~ht_cache_page();
/* new */
	virtual int	read(FILEOFS offset, byte *buf, dword len);
	virtual int	writebyte(byte b, FILEOFS offset);
	virtual int	readbyte(byte *b, FILEOFS offset);
};

/*
 *	CLASS ht_cached_file
 */

/* CF_CACHE_READS
	Allow read-caching, memory used is limited by maxcachesize */
#define CF_CACHE_READS			1

/* CF_CACHE_WRITES
	Allow write-caching, memory used is limited by maxcachesize
	   (exception see below) */
#define CF_CACHE_WRITES			2

/* CF_ONLY_EXPLICIT_FLUSHES
	Disables memory limit "maxcachesize" for write access !
	You MUST include CF_CACHE_WRITES if specifing this flag ! */
#define CF_ONLY_EXPLICIT_FLUSHES	4

class ht_tree;

#define FCNTL_CACHE_INVD	0x00000001
#define FCNTL_CACHE_FLUSH	0x00000002
#define FCNTL_CACHE_ISDIRTY	0x00000003	// +offset +range
#define FCNTL_CACHE_SET_MODE	0x00000004
#define FCNTL_CACHE_GET_MODE	0x00000005
#define FCNTL_CACHE_IS_DIRTY	0x00000006

#define ISDIRTY_SINGLEBIT	0x80000000

#define CM_READ_CACHED		1

class ht_cached_file: public ht_layer_streamfile {
protected:
	ht_tree *cache;
	dword maxcachesize;
	dword flags;
	dword alloc_index;
	FILEOFS offset;
	dword size;

	UINT cache_mode;

	dword dirty_pages;

	dword page_granularity;
	dword page_mask;
/* new */
	virtual void	cache_create();
	virtual void	cache_destroy();

		   void	cache_flush();
		   void	cache_invd();
			 
	virtual ht_cache_page *cache_page_create(FILEOFS offset);
	virtual void	cache_page_destroy(ht_cache_page *c);
	virtual ht_cache_page *cache_page_find(FILEOFS offset);
	virtual void	cache_page_flush(ht_cache_page *c);
	virtual UINT	cache_page_size();

		   void	discard_least_used_page();
		   int	writebyte(FILEOFS offset, byte b);
		   int	readbyte(FILEOFS offset, byte *b);
		   ht_cache_page *read_miss(FILEOFS offset);
		   void	set_cache_mode(UINT mode);
public:
		   void	init(ht_streamfile *streamfile, int own_streamfile, dword maxcachesize, int flags);
		   void	init(ht_streamfile *streamfile, int own_streamfile, dword maxcachesize, int flags, dword page_granularity);
	virtual void	done();
/* overwritten */
	virtual UINT	get_size();
	virtual UINT	read(void *buf, UINT size);
	virtual int	seek(FILEOFS offset);
	virtual FILEOFS tell();
	virtual int	vcntl(UINT cmd, va_list vargs);
	virtual UINT	write(void *buf, UINT size);
};

/*
 *	CLASS ht_cache_page_pbd
 */

/* pbd = per-bit write dirtyness */

class ht_cache_page_pbd: public ht_cache_page {
public:
	byte *writedirty;
	ht_streamfile *file;

	ht_cache_page_pbd(FILEOFS offset, dword size, dword alloc_index, ht_streamfile *file);
	~ht_cache_page_pbd();
/* new */
	bool isdirty(FILEOFS offset, UINT range);
	void cleanup();
};

/*
 *	CLASS ht_cached_file_pbd
 */

class ht_cached_file_pbd: public ht_cached_file {
protected:
/* overwritten */
		   bool	cache_isdirty(FILEOFS offset, UINT range);
	virtual ht_cache_page *cache_page_create(FILEOFS offset);
	virtual void	cache_page_flush(ht_cache_page *c);
	virtual UINT	cache_page_size();
public:
		   void	init(ht_streamfile *streamfile, int own_streamfile, dword maxcachesize, int flags);
		   void	init(ht_streamfile *streamfile, int own_streamfile, dword maxcachesize, int flags, dword page_granularity);
/* overwritten */
	virtual int	vcntl(UINT cmd, va_list vargs);
};

#endif /* __CSTREAM_H__ */
